from rocolib.api.components import FoldedComponent
from rocolib.api.composables.graph.Face import Rectangle as Shape
from rocolib.api.ports import AnchorPort
from rocolib.utils.utils import decorateGraph
from rocolib.utils.transforms import Translate, RotateZ
from rocolib.utils.numsym import dot, deg2rad


class ServoMotor(FoldedComponent):
    def define(self):
        self.addParameter("angle", 0, paramType="angle", minValue=None, maxValue=None)
        self.addParameter("servo", "fs90r", values=ServoMotor.dims)
        self.addInterface("mount", AnchorPort(self, self.getGraph(), "horn", Translate([0,0,0])))
        self.addFaceInterface("horn", "horn")

    def assemble(self):
        s = self.p.servo

        dz = s.get("hornheight")
        dy = s.get("motorlength") / 2 - s.get("hornoffset")

        f = Shape("horn", 0, 0)
        decorateGraph(f, Shape("hole", 1, 1))
        self.addFace(f)
        self.setInterface("mount", AnchorPort(self, self.getGraph(), "horn", dot(RotateZ(deg2rad(self.p.angle)), Translate([0,-dy,dz]))))


    '''
    Servo dimension parameters:

                 |<-G->|
    ^      =====v===== { H
    E          _I_
    v ________| | |_____
      ^ |       |<-F->|<> D
      | |             |
      B | <--- A ---> |
      | | (X) C       |
      v |_____________|

    A : motorlength
    B : motorheight
    C : motorwidth
    D : shoulderlength

    E : hornheight
    F : hornoffset

    G : hornlength
    H : horndepth

    '''

    dims = dict(
        ds2g = dict(
            motorlength   = 17,
            motorwidth    = 8.5,
            motorheight   = 11,
            shoulderlength= 3.5,
            hornlength    = 11,
            hornheight    = 9,
            hornoffset    = 4,
            horndepth     = 1,
        ),
        s4303r = dict(
            motorlength   = 31,
            motorwidth    = 17,
            motorheight   = 29,
            shoulderlength= 10,
            hornlength    = 38,
            hornheight    = 14,
            hornoffset    = 7,
            horndepth     = 2,
        ),
        tgy1370a = dict(
            motorlength   = 20,
            motorwidth    = 9,
            motorheight   = 14,
            shoulderlength= 4,
            hornlength    = 7,
            hornheight    = 10,
            hornoffset    = 4,
            horndepth     = 2,
        ),
        fs90r = dict(
            motorlength   = 23,
            motorwidth    = 13,
            motorheight   = 19,
            shoulderlength= 5,
            hornlength    = 10,
            hornheight    = 16,
            hornoffset    = 8,
            horndepth     = 2,
        ),
    )


if __name__ == "__main__":
    ServoMotor.test()
    print(ServoMotor.dims)
