import networkx as nx
import os
from os.path import dirname, realpath, basename, join
from glob import glob
import importlib
import logging
from yaml import safe_load
import sys
if sys.version_info[:2] >= (3, 10):
    # pylint: disable=no-name-in-module
    from importlib.metadata import entry_points
else:
    from importlib_metadata import entry_points
from platformdirs import user_data_dir
from collections import OrderedDict


log = logging.getLogger(__name__)
ROCOLIB_LIBRARY = dirname(realpath(__file__))
USER_LIBRARY = user_data_dir("rocolib", "rocolib")

def _makeLibDirs():
    dirs = OrderedDict(user = USER_LIBRARY)
    mods = OrderedDict()
    libs = entry_points(group="rocolib.libraries")
    for lib in libs.names:
        libmodule = libs[lib]
        dirs[lib] = libmodule.load().ROCOLIB_LIBRARY
        mods[lib] = libmodule.value
    dirs["rocolib"] = ROCOLIB_LIBRARY
    mods["rocolib"] = "rocolib.library"
    return dirs, mods

libDirs, libModules = _makeLibDirs()

def _makeLibItems(ext):
    data = {}
    for k, d in libDirs.items():
        files = [ basename(f)[:-len(ext)-1] for f in glob(d + f"/[!_]*.{ext}")]
        for f in files:
            fullname = f"{k}:{f}"
            # XXX TODO: what if a pyComponent in one library is a yamlComponent in another or vice versa?
            if f not in data:
                data[f] = fullname
            else:
                data[fullname] = fullname
    return data

yamlComponents = _makeLibItems("yaml")
pyComponents = _makeLibItems("py")

allComponents = {**pyComponents, **yamlComponents}

def _load_yaml(c):
    with open(getYamlFile(c), 'r') as fd:
        return safe_load(fd)

def _getSubcomponents(c):
    try:
        return set((x['classname'] for x in _load_yaml(c).get('subcomponents', dict()).values()))
    except (ValueError, FileNotFoundError):
        return set()

def _getComponentNx():
    net = nx.DiGraph()
    net.add_nodes_from(allComponents.keys())
    for c in allComponents:
        for sc in _getSubcomponents(c):
            if sc in allComponents:
                net.add_edge(sc, c)
            else:
                for k, v in allComponents.items():
                    if sc == v:
                        net.add_edge(k, c)
                        break
                else:
                    log.warning(f"Subcomponent {sc} of component {c} not found, ignoring")
    return net

def getComponentTree():
    return nx.topological_generations(_getComponentNx())

def getYamlFile(c):
    if c in yamlComponents:
        c = yamlComponents[c]
    elif ":" not in c:
        raise ValueError(f"Cannot find appropriate yaml file for {c}")
    d, c = c.split(":")
    return f"{libDirs[d]}/{c}.yaml"

def getPyComponent(c):
    if c in pyComponents:
        c = pyComponents[c]
    elif ":" not in c:
        raise ValueError(f"Cannot find appropriate python file / plugin library for {c}")
    d, c = c.split(":")

    obj = getattr(importlib.import_module(f"{libModules[d]}.{c}"), c)
    return obj()

def getComponent(c, **kwargs):
    '''
    Here we are doing Dynamic instantiation from string name of a class in dynamically imported module
    Parameter c (str): component name e.g. 'Stool'
    kwargs: component parameter / value pairs (possibly including name)
    '''
    from rocolib.api.components import newComponent
    if c in pyComponents.keys() or c in pyComponents.values():
        log.debug("Found python component in library, instantiating and returning object")
        my_obj = getPyComponent(c)
    elif c in yamlComponents.keys() or c in yamlComponents.values():
        log.debug("Found yaml component in library, creating newComponent from yamlFile")
        my_obj = newComponent(c, yamlFile=getYamlFile(c))
    else:
        raise ValueError(f"Component {c} not found in library")

    for k, v in kwargs.items():
        if k == 'name':
            my_obj.name = v
        else:
            my_obj.setParameter(k, v)
    if 'name' not in kwargs:
        my_obj.name = c

    return my_obj

def save(c, lib=None):
    if lib is True:
        d = ROCOLIB_LIBRARY
    elif lib:
        d = libDirs[lib]
    else:
        d = USER_LIBRARY
    os.makedirs(d, exist_ok=True)

    # XXX TODO: Check for collisions?
    # if collision:
    #   flag to allow rebuilding, and fail otherwise?
    #   if no flag, check if source matches and rebuild if so, fail otherwise?
    return c.toYaml(d, c.cName + ".yaml", reldir=ROCOLIB_LIBRARY)

def rebuild(built=None):
    if built is None:
        built = set()
    success = True
    for c in yamlComponents:
        success &= rebuildComponent(c, built, throw=False)
    assert success, "Error rebuilding all components, check stdout/stderr for details"

def rebuildComponent(c, built=None, throw=True):
    if c not in yamlComponents:
        log.debug(f"{c} is not a yaml Component, skipping")
        return True
    if built is None:
        built = set()
    if c in built:
        log.debug(f"{c} has been rebuilt, skipping")
        return True

    definition = _load_yaml(c)
    src = definition.get("source", None)
    success = True
    if src:
        log.info(f"Rebuilding {c} from {ROCOLIB_LIBRARY}/{src}...")

        subcomponents = set()
        for k, v in definition.get("subcomponents", dict()).items():
            subcomponents.add(v["classname"])
        for sc in subcomponents:
            rebuildComponent(sc, built)

        # XXX TODO: Test to make sure we don't call this script and then infinitely recurse!
        log.debug(f"Calling os.system: % python {ROCOLIB_LIBRARY}/{src}")
        if os.system(f"python {ROCOLIB_LIBRARY}/{src}"):
            success = False

        built.add(c)
        log.debug(repr(built))
        log.debug(f"Done rebuilding {c}.")
    if throw:
        assert success, f"Error rebuilding {c}, check stdout/stderr for details"
    return success
